"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
const log = require("../util/logger");
const zlib = require("zlib");
const async = require("async");
const transform_bolt_1 = require("./transform_bolt");
/////////////////////////////////////////////////////////////////////////////
const injection_placeholder = "##INJECT##";
const injection_placeholder_field = "##INJECT2##";
/** This bolt writes incoming messages to file. */
class FileAppendBolt {
    constructor() {
        this.name = null;
        this.log_prefix = null;
        this.current_data = new Map();
        this.split_value = new Set();
        this.current_file_contains_data = false;
        this.split_by_field = null;
        this.propagate_errors = true;
    }
    init(name, config, context, callback) {
        try {
            this.name = name;
            this.log_prefix = `[FileAppendBolt ${this.name}] `;
            this.file_name_template = config.file_name_template;
            this.prepend_timestamp = config.prepend_timestamp;
            this.split_over_time = config.split_over_time;
            this.split_period = config.split_period || 60 * 60 * 1000;
            if (config.split_by_field) {
                this.split_by_field = config.split_by_field.split(".");
            }
            this.compress = config.compress;
            this.propagate_errors = (config.propagate_errors == undefined) ? true : config.propagate_errors;
            // prepare filename template for injection
            if (this.split_over_time) {
                const ext = path.extname(this.file_name_template);
                this.next_split_after = Math.floor(Date.now() / this.split_period) * this.split_period;
                this.file_name_template =
                    this.file_name_template.slice(0, this.file_name_template.length - ext.length) +
                        "_" + injection_placeholder +
                        (this.split_by_field ? "_" + injection_placeholder_field : "") +
                        ext;
            }
            else {
                this.file_name_current = this.file_name_template;
                if (config.delete_existing) {
                    if (fs.existsSync(this.file_name_current)) {
                        fs.unlinkSync(this.file_name_current);
                    }
                }
            }
            callback();
        }
        catch (e) {
            callback(e);
        }
    }
    heartbeat() {
        this.writeToFile(() => {
            // no-op
        });
    }
    shutdown(callback) {
        const cb = err => {
            if (err) {
                log.logger().error(this.log_prefix + "Error in shutdown");
                log.logger().exception(err);
            }
            if (err && this.propagate_errors) {
                return callback(err);
            }
            else {
                return callback();
            }
        };
        this.writeToFile(err => {
            if (err) {
                return cb(err);
            }
            if (this.compress && this.current_file_contains_data) {
                this.zipCurrentFile(cb);
            }
            else {
                cb(null);
            }
        });
    }
    receive(data, stream_id, callback) {
        let s = "";
        try {
            if (this.prepend_timestamp) {
                s += this.toISOFormatLocal(Date.now()) + " ";
            }
            s += JSON.stringify(data);
            let key = "";
            if (this.split_by_field) {
                let obj = data;
                for (let i = 0; i < this.split_by_field.length - 1; i++) {
                    obj = obj[this.split_by_field[i]];
                }
                key = obj[this.split_by_field[this.split_by_field.length - 1]];
            }
            if (!this.current_data.has(key)) {
                this.current_data.set(key, []);
            }
            this.current_data.get(key).push(s + "\n");
            callback();
        }
        catch (e) {
            log.logger().error(this.log_prefix + "Error in receive");
            log.logger().exception(e);
            if (this.propagate_errors) {
                return callback(e);
            }
            else {
                return callback();
            }
        }
    }
    toISOFormatLocal(d) {
        const tzoffset = (new Date()).getTimezoneOffset() * 60000; // offset in milliseconds
        const s = (new Date(d - tzoffset)).toISOString().slice(0, -1);
        return s;
    }
    fileNameTimestampValue() {
        const d = Math.floor(Date.now() / this.split_period) * this.split_period;
        let s = this.toISOFormatLocal(d);
        s = s.slice(0, s.indexOf("."));
        s = s.replace(/\:/ig, "_").replace(/\-/ig, "_");
        return s;
    }
    writeToFile(callback) {
        if (this.current_data.size == 0) {
            return callback();
        }
        const d = Date.now();
        const do_file_split = (this.split_over_time && this.next_split_after < d);
        async.series([
            xcallback => {
                if (!do_file_split) {
                    return xcallback();
                }
                // perform compression of existing file if it exists.
                // only used when file split will occur
                // otherwise we just zip at shutdown.
                this.zipCurrentFile(xcallback);
            },
            xcallback => {
                if (!do_file_split) {
                    return xcallback();
                }
                // calculate new file name
                this.current_file_contains_data = false;
                this.file_name_current = this.file_name_template.replace(injection_placeholder, this.fileNameTimestampValue());
                log.logger().debug(`${this.log_prefix} new file generated: ${this.file_name_current}`);
                this.next_split_after = d + this.split_period;
                xcallback();
            },
            xcallback => {
                // write data to current file
                this.current_data.forEach((value, key) => {
                    const lines = value;
                    this.split_value.add(key);
                    const fname = this.file_name_current.replace(injection_placeholder_field, key);
                    for (const line of lines) {
                        fs.appendFileSync(fname, line);
                    }
                });
                this.current_data.clear();
                this.current_file_contains_data = true;
                xcallback();
            }
        ], callback);
    }
    /** Zip current file if it exists  */
    zipCurrentFile(callback) {
        if (this.compress && this.current_file_contains_data) {
            const fnames = [];
            this.split_value.forEach((value, key) => {
                const fname = this.file_name_current.replace(injection_placeholder_field, key);
                fnames.push(fname);
            });
            async.eachLimit(fnames, 3, (item, xcallback) => {
                if (fs.existsSync(item)) {
                    log.logger().debug(`${this.log_prefix} compressing current file: ${item}`);
                    this.zipFile(item, xcallback);
                }
                else {
                    xcallback();
                }
            }, callback);
        }
        else {
            callback();
        }
    }
    /** Perform low-level zipping */
    zipFile(fname, callback) {
        const filePath = path.resolve(fname);
        if (!fs.existsSync(filePath)) {
            return callback(new Error("Cannot zip, filename is missing: " + filePath));
        }
        let counter = 0;
        let gzFilePath = path.resolve(fname + "_" + counter + ".gz");
        while (fs.existsSync(gzFilePath)) {
            counter++;
            gzFilePath = path.resolve(fname + "_" + counter + ".gz");
        }
        try {
            const gzOption = {
                level: zlib.Z_BEST_SPEED,
                memLevel: zlib.Z_BEST_SPEED
            };
            const gzip = zlib.createGzip(gzOption);
            const inputStream = fs.createReadStream(filePath);
            const outStream = fs.createWriteStream(gzFilePath);
            inputStream.pipe(gzip).pipe(outStream);
            outStream.on("finish", err => {
                if (err) {
                    return callback(err);
                }
                fs.unlink(filePath, callback);
            });
        }
        catch (e) {
            if (fs.existsSync(gzFilePath)) {
                fs.unlinkSync(gzFilePath);
            }
            callback(e);
        }
    }
}
exports.FileAppendBolt = FileAppendBolt;
/** This bolt writes incoming messages to file in CSV format. */
class CsvFileAppendBolt {
    constructor() {
        this.file_name = null;
        this.transform = null;
        this.direct_fields = null;
        this.current_data = [];
    }
    init(_name, config, _context, callback) {
        try {
            this.file_name = config.file_name;
            this.delimiter = config.delimiter || ",";
            if (config.delete_existing) {
                if (fs.existsSync(this.file_name)) {
                    fs.unlinkSync(this.file_name);
                }
            }
            if (config.header) {
                fs.appendFileSync(this.file_name, config.header + "\n");
            }
            if (config.fields) {
                const fields = config.fields;
                this.prepareTransform(fields);
            }
            callback();
        }
        catch (e) {
            callback(e);
        }
    }
    heartbeat() {
        this.writeToFile(() => {
            // no-op
        });
    }
    shutdown(callback) {
        this.writeToFile(callback);
    }
    receive(data, stream_id, callback) {
        try {
            // transform data into CSV
            data = JSON.parse(JSON.stringify(data));
            // if output fields were not defined
            if (!this.transform) {
                // there is no transformation, export all fields
                // if this is the first received record, collect the fields
                if (!this.direct_fields) {
                    this.direct_fields = Object.keys(data).sort();
                    this.current_data.push(this.direct_fields.join(this.delimiter));
                }
                const line = this.direct_fields
                    .map(x => "" + data[x])
                    .join(this.delimiter);
                this.current_data.push(line);
            }
            else {
                // perform the transformation
                const result = this.transform.transform(data);
                const line = Object.keys(result)
                    .map(x => "" + result[x])
                    .join(this.delimiter);
                this.current_data.push(line);
            }
            callback();
        }
        catch (e) {
            log.logger().error("Error in receive");
            log.logger().exception(e);
            return callback(e);
        }
    }
    writeToFile(callback) {
        if (this.current_data.length == 0) {
            return callback();
        }
        const data = this.current_data;
        this.current_data = [];
        const lines = data.join("\n") + "\n";
        fs.appendFile(this.file_name, lines, callback);
    }
    prepareTransform(fields) {
        const transform_template = {};
        for (let i = 0; i < fields.length; i++) {
            transform_template["field_" + i] = fields[i];
        }
        this.transform = new transform_bolt_1.TransformHelper(transform_template);
    }
}
exports.CsvFileAppendBolt = CsvFileAppendBolt;
//# sourceMappingURL=file_append_bolt.js.map