"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const async = require("async");
const path = require("path");
const fb = require("./std_nodes/filter_bolt");
const pb = require("./std_nodes/post_bolt");
const cb = require("./std_nodes/console_bolt");
const ab = require("./std_nodes/attacher_bolt");
const ac = require("./std_nodes/accumulator_bolt");
const tb = require("./std_nodes/transform_bolt");
const gb = require("./std_nodes/get_bolt");
const rb = require("./std_nodes/router_bolt");
const bb = require("./std_nodes/bomb_bolt");
const fab = require("./std_nodes/file_append_bolt");
const fab2 = require("./std_nodes/file_append_bolt_ex");
const cntb = require("./std_nodes/counter_bolt");
const ttb = require("./std_nodes/type_transform_bolt");
const prb = require("./std_nodes/process_bolt");
const frs = require("./std_nodes/file_reader_spout");
const ps = require("./std_nodes/process_spout");
const rs = require("./std_nodes/rest_spout");
const ts = require("./std_nodes/timer_spout");
const gs = require("./std_nodes/get_spout");
const rss = require("./std_nodes/rss_spout");
const tss = require("./std_nodes/test_spout");
const ds = require("./std_nodes/dir_watcher_spout");
const tel = require("./util/telemetry");
const log = require("./util/logger");
const topology_async_wrappers_1 = require("./topology_async_wrappers");
const forward_bolt_1 = require("./std_nodes/forward_bolt");
const NEXT_SLEEP_TIMEOUT = 1 * 1000; // number of miliseconds to "sleep" when spout.next() returned no data
/** Base class for spouts and bolts - contains telemetry support */
class TopologyNodeBase {
    constructor(name, telemetry_timeout) {
        this.name = name;
        this.telemetry = new tel.Telemetry(name);
        this.telemetry_total = new tel.Telemetry(name);
        this.telemetry_next_emit = Date.now();
        this.telemetry_timeout = telemetry_timeout || 60 * 1000;
        this.firstErrorMessage = "";
    }
    /** This method checks if telemetry data should be emitted
     * and calls provided callback if that is the case.
     */
    telemetryHeartbeat(emitCallback) {
        const now = Date.now();
        if (now >= this.telemetry_next_emit) {
            const msg = {
                last: this.telemetry.get(),
                name: this.name,
                total: this.telemetry_total.get(),
                ts: Date.now()
            };
            emitCallback(msg, "$telemetry");
            this.telemetry.reset();
            this.telemetry_next_emit = now + this.telemetry_timeout;
        }
    }
    /** Adds duration to internal telemetry */
    telemetryAdd(duration) {
        this.telemetry.add(duration);
        this.telemetry_total.add(duration);
    }
    /** helper function that sets isError flag when a callback is called with an error */
    wrapCallbackSetError(callback) {
        return (err) => {
            if (err) {
                this.isError = true;
                if (this.firstErrorMessage == "") {
                    this.firstErrorMessage = err.message;
                }
                log.logger().error("Internal error in " + this.name + ". Setting error flag.");
                log.logger().exception(err);
            }
            try {
                callback(err);
            }
            catch (e) {
                log.logger().error("THIS SHOULD NOT HAPPEN: exception THROWN in callback!");
                log.logger().exception(e);
            }
        };
    }
}
exports.TopologyNodeBase = TopologyNodeBase;
/** Wrapper for spout */
class TopologySpoutWrapper extends TopologyNodeBase {
    /** Factory method for sys spouts */
    static createSysSpout(spout_config) {
        switch (spout_config.cmd) {
            case "timer": return new ts.TimerSpout();
            case "get": return new gs.GetSpout();
            case "rest": return new rs.RestSpout();
            case "dir": return new ds.DirWatcherSpout();
            case "file_reader": return new frs.FileReaderSpout();
            case "process": return new ps.ProcessSpout();
            case "process-continuous": return new ps.ProcessSpoutContinuous();
            case "rss": return new rss.RssSpout();
            case "test": return new tss.TestSpout();
            default: throw new Error("Unknown sys spout type: " + spout_config.cmd);
        }
    }
    /** Constructor needs to receive all data */
    constructor(config, context) {
        super(config.name, config.telemetry_timeout);
        this.name = config.name;
        this.context = context;
        this.working_dir = config.working_dir;
        this.cmd = config.cmd;
        this.subtype = config.subtype;
        this.init_params = config.init || {};
        this.isError = false;
        try {
            if (config.type == "sys") {
                this.spout = createSysSpout(config);
            }
            else if (config.type == "module_class") {
                this.makeWorkingDirAbsolute();
                const module = require(this.working_dir);
                const obj = new module[this.cmd](this.subtype);
                if (!obj) {
                    throw new Error(`Spout factory returned null: ${this.working_dir}, cmd=${this.cmd}, subtype=${this.subtype}`);
                }
                if (obj.init.length < 4) {
                    this.spout = new topology_async_wrappers_1.SpoutAsyncWrapper(obj);
                }
                else {
                    this.spout = obj;
                }
            }
            else if (config.type == "module_method") {
                this.makeWorkingDirAbsolute();
                const module = require(this.working_dir);
                const obj = module[this.cmd](this.subtype);
                if (!obj) {
                    throw new Error(`Spout factory returned null: ${this.working_dir}, cmd=${this.cmd}, subtype=${this.subtype}`);
                }
                if (obj.init.length < 4) {
                    this.spout = new topology_async_wrappers_1.SpoutAsyncWrapper(obj);
                }
                else {
                    this.spout = obj;
                }
            }
            else {
                this.working_dir = path.resolve(this.working_dir); // path may be relative to current working dir
                const module_path = path.join(this.working_dir, this.cmd);
                const obj = require(module_path).create(this.subtype);
                if (!obj) {
                    throw new Error(`Spout factory returned null: ${module_path}, subtype=${this.subtype}`);
                }
                if (obj.init.length < 4) {
                    this.spout = new topology_async_wrappers_1.SpoutAsyncWrapper(obj);
                }
                else {
                    this.spout = obj;
                }
            }
        }
        catch (e) {
            log.logger().error(`Error while creating an inproc spout (${this.name})`);
            log.logger().exception(e);
            throw e;
        }
        this.emitCallback = (data, stream_id, callback) => {
            config.onEmit(data, stream_id, callback);
        };
        this.errorCallback = config.onError || (() => {
            // no-op
        });
        this.errorCallback = this.wrapCallbackSetError(this.errorCallback);
        this.isPaused = true;
        this.isShuttingDown = false;
        this.nextTs = Date.now();
    }
    /** Returns name of this node */
    getName() {
        return this.name;
    }
    /** Returns inner spout object */
    getSpoutObject() {
        return this.spout;
    }
    /** Handler for heartbeat signal */
    heartbeat() {
        if (this.isError) {
            return;
        }
        if (this.isPaused) {
            return;
        }
        try {
            this.spout.heartbeat();
        }
        catch (e) {
            log.logger().error(`Error in spout (${this.name}) heartbeat`);
            log.logger().exception(e);
            return this.errorCallback(e);
        }
        this.telemetryHeartbeat((msg, stream_id) => {
            this.emitCallback(msg, stream_id, () => {
                // no-op
            });
        });
    }
    /** Shuts down the process */
    shutdown(callback) {
        // wrap callback to set self.isError when an exception passed
        callback = this.wrapCallbackSetError(callback);
        if (this.isError) {
            return callback(new Error(`Spout (${this.name}) has error flag set. ` +
                `Shutdown refused. Last error: ${this.firstErrorMessage}`));
        }
        // without an exception the caller will think that everything shut down nicely already
        // when we call shutdown twice by mistake
        if (this.isShuttingDown) {
            return callback(new Error(`Spout (${this.name}) is already shutting down.`));
        }
        this.isShuttingDown = true;
        try {
            this.spout.shutdown(callback);
        }
        catch (e) {
            log.logger().error(`Unhandled error in spout (${this.name}) shutdown`);
            log.logger().exception(e);
            return callback(e);
        }
    }
    /** Initializes child object. */
    init(callback) {
        // wrap callback to set self.isError when an exception passed
        callback = this.wrapCallbackSetError(callback);
        if (this.isError) {
            return callback(new Error(`Spout (${this.name}) has error flag set. ` +
                `Init refused. Last error: ${this.firstErrorMessage}`));
        }
        if (this.initCalled) {
            return callback(new Error(`Init already called in spout (${this.name})`));
        }
        this.initCalled = true;
        try {
            this.spout.init(this.name, this.init_params, this.context, callback);
        }
        catch (e) {
            log.logger().error(`Unhandled error in spout (${this.name}) init`);
            log.logger().exception(e);
            callback(e);
        }
    }
    /** Sends run signal and starts the "pump" */
    run() {
        if (!this.isPaused) {
            return; // already running
        }
        this.isPaused = false;
        try {
            this.spout.run();
        }
        catch (e) {
            log.logger().error(`Error in spout (${this.name}) run`);
            log.logger().exception(e);
            this.errorCallback(e);
            return;
        }
        async.whilst(() => !this.isPaused && !this.isError, xcallback => {
            if (Date.now() < this.nextTs) {
                const sleep = this.nextTs - Date.now();
                setTimeout(() => { xcallback(); }, sleep);
            }
            else {
                this.next(xcallback);
            }
        }, (err) => {
            if (err) {
                log.logger().error(`Error in spout (${this.name}) next`);
                log.logger().exception(err);
                this.errorCallback(err);
            }
        });
    }
    /** Sends pause signal to child */
    pause() {
        if (this.isPaused) {
            return; // already paused
        }
        this.isPaused = true;
        try {
            this.spout.pause();
        }
        catch (e) {
            log.logger().error(`Error in spout (${this.name}) pause`);
            log.logger().exception(e);
            this.errorCallback(e);
        }
    }
    /** Requests next data message */
    next(callback) {
        // wrap callback to set self.isError when an exception passed
        callback = this.wrapCallbackSetError(callback);
        if (this.isPaused || this.isError) {
            // run will stop calling next when isError is true
            return callback();
        }
        else {
            const ts_start = Date.now();
            setImmediate(() => {
                try {
                    this.spout.next((err, data, stream_id) => {
                        this.telemetryAdd(Date.now() - ts_start);
                        if (err) {
                            log.logger().exception(err);
                            return callback(err);
                        }
                        if (!data) {
                            this.nextTs = Date.now() + NEXT_SLEEP_TIMEOUT;
                            return callback();
                        }
                        else {
                            try {
                                this.emitCallback(data, stream_id, callback);
                            }
                            catch (e) {
                                // there was an error, don't call the child's xcallback
                                return callback(e);
                            }
                        }
                    });
                }
                catch (e) {
                    return callback(e);
                }
            });
        }
    }
    /** Utility function for making working dir absolute - used to avoid some problematic situations */
    makeWorkingDirAbsolute() {
        if (!path.isAbsolute(this.working_dir)) {
            this.working_dir = path.resolve(this.working_dir);
        }
    }
}
exports.TopologySpoutWrapper = TopologySpoutWrapper;
/** Wrapper for bolt */
class TopologyBoltWrapper extends TopologyNodeBase {
    /** Factory method for sys bolts */
    static createSysBolt(bolt_config) {
        switch (bolt_config.cmd) {
            case "console": return new cb.ConsoleBolt();
            case "filter": return new fb.FilterBolt();
            case "forward": return new forward_bolt_1.ForwardrBolt();
            case "attacher": return new ab.AttacherBolt();
            case "accumulator": return new ac.AccumulatorBolt();
            case "transform": return new tb.TransformBolt();
            case "post": return new pb.PostBolt();
            case "process": return new prb.ProcessBoltContinuous();
            case "get": return new gb.GetBolt();
            case "router": return new rb.RouterBolt();
            case "file_append_csv": return new fab.CsvFileAppendBolt();
            case "file_append": return new fab.FileAppendBolt();
            case "file_append_ex": return new fab2.FileAppendBoltEx();
            case "date_transform": return new ttb.TypeTransformBolt();
            case "type_transform": return new ttb.TypeTransformBolt();
            case "date2numeric_transform": return new ttb.DateToNumericTransformBolt();
            case "bomb": return new bb.BombBolt();
            case "counter": return new cntb.CounterBolt();
            default: throw new Error("Unknown sys bolt type: " + bolt_config.cmd);
        }
    }
    /** Constructor needs to receive all data */
    constructor(config, context) {
        super(config.name, config.telemetry_timeout);
        this.name = config.name;
        this.context = context;
        this.working_dir = config.working_dir;
        this.cmd = config.cmd;
        this.subtype = config.subtype;
        this.isError = false;
        this.init_params = config.init || {};
        this.init_params.onEmit = (data, stream_id, callback) => {
            config.onEmit(data, stream_id, callback);
        };
        this.emitCallback = this.init_params.onEmit;
        this.errorCallback = config.onError || (() => {
            // no-op
        });
        this.errorCallback = this.wrapCallbackSetError(this.errorCallback);
        this.allow_parallel = config.allow_parallel || false;
        this.isShuttingDown = false;
        this.inSend = 0;
        this.pendingSendRequests = [];
        this.pendingShutdownCallback = null;
        try {
            if (config.type == "sys") {
                this.bolt = createSysBolt(config);
            }
            else if (config.type == "module_class") {
                this.makeWorkingDirAbsolute();
                const module = require(this.working_dir);
                const obj = new module[this.cmd](this.subtype);
                if (!obj) {
                    throw new Error(`Bolt factory returned null: ${this.working_dir}, cmd=${this.cmd}, subtype=${this.subtype}`);
                }
                if (obj.init.length < 4) {
                    this.bolt = new topology_async_wrappers_1.BoltAsyncWrapper(obj);
                }
                else {
                    this.bolt = obj;
                }
            }
            else if (config.type == "module_method") {
                this.makeWorkingDirAbsolute();
                const module = require(this.working_dir);
                const obj = module[this.cmd](this.subtype);
                if (!obj) {
                    throw new Error(`Bolt factory returned null: ${this.working_dir}, cmd=${this.cmd}, subtype=${this.subtype}`);
                }
                if (obj.init.length < 4) {
                    this.bolt = new topology_async_wrappers_1.BoltAsyncWrapper(obj);
                }
                else {
                    this.bolt = obj;
                }
            }
            else {
                this.working_dir = path.resolve(this.working_dir); // path may be relative to current working dir
                const module_path = path.join(this.working_dir, this.cmd);
                const obj = require(module_path).create(this.subtype);
                if (!obj) {
                    throw new Error(`Bolt factory returned null: ${module_path}, subtype=${this.subtype}`);
                }
                if (obj.init.length < 4) {
                    this.bolt = new topology_async_wrappers_1.BoltAsyncWrapper(obj);
                }
                else {
                    this.bolt = obj;
                }
            }
        }
        catch (e) {
            log.logger().error(`Error while creating an inproc bolt (${this.name})`);
            log.logger().exception(e);
            throw e;
        }
    }
    /** Returns name of this node */
    getName() {
        return this.name;
    }
    /** Returns inner bolt object */
    getBoltObject() {
        return this.bolt;
    }
    /** Handler for heartbeat signal */
    heartbeat() {
        const self = this;
        if (self.isError) {
            return;
        }
        try {
            self.bolt.heartbeat();
        }
        catch (e) {
            log.logger().error(`Error in bolt (${this.name}) heartbeat`);
            log.logger().exception(e);
            self.errorCallback(e);
            return;
        }
        self.telemetryHeartbeat((msg, stream_id) => {
            self.emitCallback(msg, stream_id, () => {
                // no-op
            });
        });
    }
    /** Shuts down the child */
    shutdown(callback) {
        // wrap callback to set self.isError when an exception passed
        callback = this.wrapCallbackSetError(callback);
        if (this.isError) {
            return callback(new Error(`Bolt (${this.name}) has error flag set. ` +
                `Shutdown refused. Last error: ${this.firstErrorMessage}`));
        }
        // without an exception the caller will think that everything shut down nicely already
        // when we call shutdown twice by mistake
        if (this.isShuttingDown) {
            return callback(new Error(`Bolt (${this.name}) is already shutting down.`));
        }
        this.isShuttingDown = true;
        try {
            if (this.inSend === 0) {
                return this.bolt.shutdown(callback);
            }
            else {
                const cb2 = () => { this.bolt.shutdown(callback); };
                this.pendingShutdownCallback = cb2;
            }
        }
        catch (e) {
            callback(e);
        }
    }
    /** Initializes child object. */
    init(callback) {
        // wrap callback to set self.isError when an exception passed
        callback = this.wrapCallbackSetError(callback);
        if (this.isError) {
            return callback(new Error(`Bolt (${this.name}) has error flag set. ` +
                `Init refused. Last error: ${this.firstErrorMessage}`));
        }
        if (this.initCalled) {
            return callback(new Error(`Bolt (${this.name}) init already called.`));
        }
        this.initCalled = true;
        try {
            this.bolt.init(this.name, this.init_params, this.context, callback);
        }
        catch (e) {
            log.logger().error(`Error in bolt (${this.name}) init`);
            log.logger().exception(e);
            callback(e);
        }
    }
    /** Sends data to child object. */
    receive(data, stream_id, callback) {
        // wrap callback to set self.isError when an exception passed
        callback = this.wrapCallbackSetError(callback);
        const self = this;
        if (self.isError) {
            return callback(new Error(`Bolt ${self.name} has error flag set. Last error: ${this.firstErrorMessage}`));
        }
        const ts_start = Date.now();
        if (self.inSend > 0 && !self.allow_parallel) {
            self.pendingSendRequests.push({ callback, data, stream_id });
        }
        else {
            self.inSend++;
            try {
                self.bolt.receive(data, stream_id, err => {
                    self.telemetryAdd(Date.now() - ts_start);
                    callback(err);
                    if (err) {
                        return;
                    } // stop processing if there was an error
                    self.inSend--;
                    if (self.inSend === 0) {
                        if (self.pendingSendRequests.length > 0) {
                            const d = self.pendingSendRequests[0];
                            self.pendingSendRequests = self.pendingSendRequests.slice(1);
                            self.receive(d.data, d.stream_id, d.callback);
                        }
                        else if (self.pendingShutdownCallback) {
                            const cb2 = self.pendingShutdownCallback;
                            self.pendingShutdownCallback = null;
                            cb2();
                        }
                    }
                });
            }
            catch (e) {
                callback(e);
            }
        }
    }
    /** Utility function for making working dir absolute - used to avoid some problematic situations */
    makeWorkingDirAbsolute() {
        if (!path.isAbsolute(this.working_dir)) {
            this.working_dir = path.resolve(this.working_dir);
        }
    }
}
exports.TopologyBoltWrapper = TopologyBoltWrapper;
function createSysSpout(config) {
    return TopologySpoutWrapper.createSysSpout(config);
}
exports.createSysSpout = createSysSpout;
function createSysBolt(config) {
    return TopologyBoltWrapper.createSysBolt(config);
}
exports.createSysBolt = createSysBolt;
//# sourceMappingURL=topology_local_inprocess.js.map